/*
    Valtra Tractors
    Copyright 2004 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "fsys.h"
#if defined(__APPLE__) && defined(__MACH__)
#include <CoreFoundation/CoreFoundation.h>
#endif

FileSystem fsys;


#ifdef HAVE_ZLIB
#ifdef WIN32
#include <windows.h>
//#define ZEXPORT WINAPI
#endif
#include <zlib.h>
#endif

#ifndef WIN32
#include <sys/types.h>
#include <dirent.h>
#define DIRSEP '/'
#else
#include <windows.h>
#define DIRSEP '\\'
#endif

#define BUFFERSIZE 1000

void fixDirSep(char *ptr, char sep = DIRSEP) {
	while (*ptr) {
		if (*ptr=='/' || *ptr=='\\') *ptr = sep;
		ptr++;
	}
}


void ZipFile::init(SDL_RWops *file) {
	if (file==NULL) return;
	char sig1[] = { 'P', 'K', 0x03, 0x04 };
	char sig2[] = { 'P', 'K', 0x01, 0x02 };
	int pos = 0;
	char buffer[30];
	while (true) {
		if (SDL_RWread(file,buffer,1,30)<30) break;
		Uint8 *ptr = (Uint8*) buffer;
		if (memcmp(ptr,sig1,4)) {
			if (memcmp(ptr,sig2,4)) break; // bad file?
			// OK, no more files
			break;
		}
		ptr += 8;
		Uint16 compression = *((Uint16*)ptr);
		ptr += 10;
		Uint32 csize = *((Uint32*)ptr);
		ptr += 4;
		Uint32 dsize = *((Uint32*)ptr);
		ptr += 4;
		Uint16 namelen = *((Uint16*)ptr);
		ptr += 2;
		Uint16 xlen = *((Uint16*)ptr);
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
		compression = SDL_Swap16(compression);
		csize = SDL_Swap32(csize);
		dsize = SDL_Swap32(dsize);
		namelen = SDL_Swap16(namelen);
		xlen = SDL_Swap16(xlen);
#endif
		char *name = (char*) malloc(namelen+1);
		if (name==NULL) return;
		if (SDL_RWread(file,name,1,namelen)<namelen) break;
		name[namelen] = '\0';
		pos += 30+namelen+xlen;
#ifndef HAVE_ZLIB
		if (compression==0)
#endif
			entries.push_back(ZipEntry(name,pos,compression,csize,dsize));
		free(name);
		pos += csize;
		if (SDL_RWseek(file,xlen+csize,SEEK_CUR)<0) break;
	}
	SDL_RWclose(file);
}

ZipFile::ZipFile(void *p, int size) {
	data = (Uint8*) p;
	init(SDL_RWFromMem(p,size));
}

ZipFile::ZipFile(const char *n) {
	data = NULL;
	name = n;
	init(SDL_RWFromFile(n,"rb"));
}

ZipFile::~ZipFile() {
}

int ZipFile::namecmp(const char *n1, const char *n2) {
	while (true) {
		char c1 = toupper(*n1);
		char c2 = toupper(*n2);
		if (c1=='\\') c1 = '/';
		if (c2=='\\') c2 = '/';
		if (c1<c2) return -1;
		else if (c1>c2) return 1;
		if (c1=='\0') return 0;
		n1++;
		n2++;
	}
}

struct rw_fp {
	FILE *fp;
	int begin, size, pos;
	bool autoclose;
};

int rw_fp_seek(SDL_RWops *ctx, int offset, int whence) {
	struct rw_fp *data = (struct rw_fp*) ctx->hidden.unknown.data1;
	switch (whence) {
	case SEEK_SET:
		data->pos = offset;
		break;
	case SEEK_CUR:
		data->pos += offset;
		break;
	case SEEK_END:
		data->pos = data->size+offset;
		break;
	default:
		return -1;
		break;
	}
	if (data->pos<0) data->pos = 0;
	if (data->pos>data->size) data->pos = data->size;
	if (fseek(data->fp,data->begin+data->pos,SEEK_SET)==0)
		return data->pos;
//	SDL_Error(SDL_EFSEEK);
	return -1;
}

int rw_fp_read(SDL_RWops *ctx, void *ptr, int size, int maxnum) {
	struct rw_fp *data = (struct rw_fp*) ctx->hidden.unknown.data1;
	int n = maxnum;
	int n2 = (data->size-data->pos)/size;
	if (n2<n) n = n2;
//	if (data->pos == data->size) return -1;
	n = fread(ptr,size,n,data->fp);
	data->pos += n*size;
//	if (n==0 && ferror(data->fp))
//		SDL_Error(SDL_EFREAD);
	return n;
}

int rw_fp_write(SDL_RWops *ctx, const void *ptr, int size, int num) {
	struct rw_fp *data = (struct rw_fp*) ctx->hidden.unknown.data1;
	int n = num;
	int n2 = (data->size-data->pos)/size;
	if (n2<n) n = n2;
//	if (data->pos == data->size) return -1;
	n = fwrite(ptr,size,n,data->fp);
	data->pos += n*size;
//	if (n==0 && ferror(data->fp))
//		SDL_Error(SDL_EFWRITE);
	return n;
}

int rw_fp_close(SDL_RWops *ctx) {
	struct rw_fp *data = (struct rw_fp*) ctx->hidden.unknown.data1;
	if (data->autoclose)
		fclose(data->fp);
	free(data);
	free(ctx);
	return 0;
}


SDL_RWops *rwFromFP(FILE *in, int size, bool autoclose = true) {
	SDL_RWops *rw = (SDL_RWops*) malloc(sizeof(SDL_RWops));
	if (rw==NULL) return NULL;
	rw->hidden.unknown.data1 = malloc(sizeof(struct rw_fp));
	if (rw->hidden.unknown.data1==NULL) {
		free(rw);
		return NULL;
	}
	struct rw_fp *data = (struct rw_fp*) rw->hidden.unknown.data1;
	data->fp = in;
	data->begin = ftell(in);
	data->size = size;
	data->pos = 0;
	data->autoclose = autoclose;
	rw->seek = rw_fp_seek;
	rw->read = rw_fp_read;
	rw->write = rw_fp_write;
	rw->close = rw_fp_close;
	return rw;
}

int (*orig_mem_close)(struct SDL_RWops *context);

int mem_close_free(struct SDL_RWops *context) {
	free(context->hidden.mem.base);
	return orig_mem_close(context);
}

SDL_RWops *rwFromMem(void *ptr, int size, bool free = false) {
	SDL_RWops *rw = SDL_RWFromMem(ptr,size);
	if (free) {
		orig_mem_close = rw->close;
		rw->close = mem_close_free;
	}
	return rw;
}

SDL_RWops *ZipFile::getFile(const char *n, bool need_end) {
	for (unsigned int i = 0; i < entries.size(); i++) {
		if (namecmp(entries[i].name.c_str(),n)) continue;
		if (entries[i].compression==0) {
			if (data)
				return SDL_RWFromMem(data+entries[i].offset,entries[i].dsize);
			FILE *in = fopen(name.c_str(),"rb");
			if (in==NULL) return NULL;
			fseek(in,entries[i].offset,SEEK_SET);
			if (!need_end)
				return SDL_RWFromFP(in,1);
			return rwFromFP(in,entries[i].dsize);
		} else {
#ifdef HAVE_ZLIB
			if (data) {
				Uint8 *buffer = (Uint8*) malloc(entries[i].dsize);
				if (buffer==NULL) return NULL;
/*				uLongf len = entries[i].dsize;
				int ret;
				if ((ret = uncompress(buffer,&len,data+entries[i].offset,entries[i].csize))!=Z_OK) {
					printf("Error %d in uncompress, %s\n",ret,zError(ret));
					free(buffer);
					return NULL;
				}
				return rwFromMem(buffer,len,true);
*/

				z_stream stream;
				stream.zalloc = Z_NULL;
				stream.zfree = Z_NULL;
				stream.opaque = Z_NULL;
				stream.next_out = buffer;
				stream.avail_out = entries[i].dsize;
				stream.next_in = (Bytef*) data+entries[i].offset;
				stream.avail_in = entries[i].csize;
				int ret = inflateInit2(&stream,-MAX_WBITS);
				if (ret!=Z_OK)
					fprintf(stderr,"Error %d in inflateInit, %s\n",ret,stream.msg);
				ret = inflate(&stream,Z_SYNC_FLUSH);
/*				if (ret!=Z_STREAM_END) {
					printf("tried to decompress %s, failed (%d, %s)\n",n,ret,stream.msg);
					fprintf(stderr,"End of compressed stream without Z_STREAM_END, aborting...\n");
					free(buffer);
					inflateEnd(&stream);
					return NULL;
				}
*/				inflateEnd(&stream);
				return rwFromMem(buffer,entries[i].dsize,true);

			} else {
				FILE *in = fopen(name.c_str(),"rb");
				if (in==NULL) return NULL;
				fseek(in,entries[i].offset,SEEK_SET);
				z_stream stream;
				stream.zalloc = Z_NULL;
				stream.zfree = Z_NULL;
				stream.opaque = Z_NULL;
				Uint8 *out = (Uint8*) malloc(entries[i].dsize);
				if (out==NULL) {
					fclose(in);
					return NULL;
				}
				stream.next_out = out;
				stream.avail_out = entries[i].dsize;
				char buffer[8192];
				unsigned int csize = entries[i].csize;
				int n = fread(buffer,1,sizeof(buffer)<csize?sizeof(buffer):csize,in);
				stream.next_in = (Bytef*) buffer;
				stream.avail_in = n;
				csize -= n;
				int ret = inflateInit2(&stream,-MAX_WBITS);
				if (ret!=Z_OK)
					fprintf(stderr,"Error %d in inflateInit, %s\n",ret,stream.msg);
				while (true) {
					ret = inflate(&stream,Z_SYNC_FLUSH);
					if (ret==Z_STREAM_END) break;
					if (ret==Z_OK) {
						if (stream.avail_in==0) {
							if (csize<=0) {
/*								fprintf(stderr,"End of compressed stream without Z_STREAM_END, aborting...\n");
								free(out);
								fclose(in);
								inflateEnd(&stream);
								return NULL;
*/
								break;
							}
							n = fread(buffer,1,sizeof(buffer)<csize?sizeof(buffer):csize,in);
							stream.next_in = (Bytef*) buffer;
							stream.avail_in = n;
							csize -= n;
						} else {
							fprintf(stderr,"inflate returned without avail_in==0, aborting...\n");
							free(out);
							fclose(in);
							inflateEnd(&stream);
							return NULL;
						}
					} else {
						fprintf(stderr,"Error %d in inflate (%s), aborting...\n",ret,stream.msg);
						free(out);
						fclose(in);
						inflateEnd(&stream);
						return NULL;
					}
				}
				fclose(in);
				inflateEnd(&stream);
				return rwFromMem(out,entries[i].dsize,true);
			}
#endif
		}
	}
	return NULL;
}

void ZipFile::listFiles(const char *dir, vector<string> &files) {
	char buffer[BUFFERSIZE];
	sprintf(buffer,"%s/",dir);
	fixDirSep(buffer,'/');
	unsigned int len = strlen(buffer);
	if (len==1) len = 0;
	for (unsigned int i = 0; i < entries.size(); i++) {
		if (len>0 && strncmp(buffer,entries[i].name.c_str(),len)) continue;
		if (entries[i].name.size()==len) continue; // skip the dir itself
		char buffer2[BUFFERSIZE*2];
		strcpy(buffer2,entries[i].name.c_str());
		char *ptr = buffer2;
		ptr += len;
		bool subdir = false;
		while (*ptr) {
			if (*ptr=='/') {
				ptr[1] = '\0';
				subdir = true;
				break;
			}
			ptr++;
		}
		ptr = buffer2+len;
		if (subdir) {
			unsigned int j;
			for (j = 0; j < files.size(); j++) {
				if (files[j]==ptr) break;
			}
			if (j==files.size()) files.push_back(ptr);
		} else {
			files.push_back(ptr);
		}
	}
}

RealFileSystem::RealFileSystem() {
	root = NULL;
}

RealFileSystem::RealFileSystem(const char *r) {
	char *ptr = (char*) malloc(strlen(r)+2);
	strcpy(ptr,r);
	int len = strlen(ptr);
	if (len==0) {
		free(ptr);
		root = NULL;
		return;
	}
	if (ptr[len-1]!='/' && ptr[len-1]!='\\') {
		ptr[len] = DIRSEP;
		ptr[len+1] = '\0';
	}
	root = ptr;
}

RealFileSystem::~RealFileSystem() {
	free((void*)root);
}


SDL_RWops *RealFileSystem::getFile(const char *n, bool need_end) {
	char buffer[BUFFERSIZE];
	if (root)
		sprintf(buffer,"%s%c%s",root,DIRSEP,n);
	else
		strcpy(buffer,n);
	fixDirSep(buffer);
	return SDL_RWFromFile(buffer,"rb");
}

void RealFileSystem::listFiles(const char *dir, vector<string> &files) {
#ifndef WIN32
	char buffer[BUFFERSIZE];
	if (root)
		sprintf(buffer,"%s/%s",root,dir);
	else
		strcpy(buffer,dir);
	fixDirSep(buffer);
	DIR *d = opendir(dir);
	if (d==NULL) return;
	struct dirent *entry;
	while ((entry=readdir(d))!=NULL) {
		files.push_back(entry->d_name);
	}
	closedir(d);
#else
	WIN32_FIND_DATA info;
	char buffer[BUFFERSIZE];
	if (root)
		sprintf(buffer,"%s\\%s\\*",root,dir);
	else
		sprintf(buffer,"%s\\*",dir);
	fixDirSep(buffer);
	HANDLE h = FindFirstFile(buffer,&info);
	if (h==INVALID_HANDLE_VALUE) return;
	while (true) {
		files.push_back(info.cFileName);
		if (!FindNextFile(h,&info)) {
			if (GetLastError()==ERROR_NO_MORE_FILES) break;
		}
	}
	FindClose(h);
#endif
}


void FileSystem::loadRes() {
#ifdef WIN32
	HMODULE hmod = GetModuleHandle(NULL); // hmod=NULL?
	HRSRC resinfo = FindResource(hmod,"DATAZIP","ZIP");
	if (resinfo==NULL) return;
	int size = SizeofResource(hmod,resinfo);
	HGLOBAL res = LoadResource(hmod,resinfo);
	void *data = LockResource(res);
	sources.push_back(new ZipFile(data,size));
#endif
}

FileSystem::FileSystem(int srcs) {
	if (srcs&FS_RES) loadRes();
	if (srcs&FS_ZIP) sources.push_back(new ZipFile("data.zip"));
#if defined(__APPLE__) && defined(__MACH__)
	char buffer[1000];
//	CFURLRef url = CFBundleCopyResourceURL(CFBundleGetMainBundle(), CFSTR("data.zip"), NULL, NULL);
	CFURLRef url = CFBundleCopyResourcesDirectoryURL(CFBundleGetMainBundle());
	CFURLGetFileSystemRepresentation(url, true, (UInt8*) buffer, sizeof(buffer));
	CFRelease(url);
	strcat(buffer, "/data.zip");
	sources.push_back(new ZipFile(buffer));
#endif
	if (srcs&FS_REAL) sources.push_back(new RealFileSystem());
}

FileSystem::~FileSystem() {
	for (unsigned int i = 0; i < sources.size(); i++)
		delete sources[i];
}

SDL_RWops *FileSystem::getFile(const char *n, bool need_end) {
	for (unsigned int i = 0; i < sources.size(); i++) {
		SDL_RWops *rw = sources[i]->getFile(n,need_end);
		if (rw) return rw;
	}
	return NULL;
}

void FileSystem::listFiles(const char *dir, vector<string> &files) {
	int len = strlen(dir);
	char buffer[BUFFERSIZE];
	// remove starting slash
	if (dir[0]=='/' || dir[0]=='\\') dir++;
	strcpy(buffer,dir);
	// remove trailing slash
	if (len>0) {
		if (buffer[len-1]=='/' || buffer[len-1]=='\\') buffer[len-1] = '\0';
	}
	for (unsigned int i = 0; i < sources.size(); i++)
		sources[i]->listFiles(buffer,files);
}



